package nl.cwi.sen1.tunit;

import junit.framework.TestCase;
import nl.cwi.sen1.tunit.execution.DefaultTestToolExecutorFactory;
import toolbus.ToolBus;
import toolbus.adapter.AbstractTool;
import aterm.ATermFactory;

/**
 * The TUnitTestCase class is a base class for testing toolbus scripts. This
 * class can start / stop the toolbus and connect / disconnect stubs from the
 * toolbus.
 * 
 * @author Bas Basten & Anton Lycklama a Nijeholt
 */
public abstract class TUnitTestCase extends TestCase {
	protected volatile ToolBus toolbus;
	protected final ATermFactory factory;
	protected final DefaultTestToolExecutorFactory defaultTestToolExecutorFactory;

	private volatile boolean errorsOccured = false;

	public TUnitTestCase() {
		super();

		factory = AbstractTool.getFactory();
		defaultTestToolExecutorFactory = new DefaultTestToolExecutorFactory(
				this);
	}

	/**
	 * Get the absolute top path of the current package.
	 */
	public String getTopSrcDir() {
		// Get the environment variable BUILDFILE that is passed
		// from the ant file that is generated from the build scripts.
		String topSrcDir = System.getenv("BUILDFILE");

		if (topSrcDir != null) {
			int pos = topSrcDir.indexOf("/meta-build.ant");
			if (pos != -1) {
				return topSrcDir.substring(0, pos);
			}
		}

		// When the unit test is not called from an ant script, we just
		// return the current working directory. (For use in Eclipse)
		return ".";
	}

	/**
	 * Starts the toolbus at a given port and executes a script. If a script
	 * uses include scripts the searchPath will be used to include the include
	 * scripts from.
	 * 
	 * @param searchPath
	 *            The directories to search for the included script files of the
	 *            specified script.
	 * @param scriptName
	 *            The script to be executed.
	 */
	protected void startToolbus(String searchPath, String scriptName){
		String[] args = { "-I" + searchPath, "-S" + scriptName };

		toolbus = new ToolBus(args);
		toolbus.setToolExecutorFactory(defaultTestToolExecutorFactory);

		try{
			toolbus.parsecup();
			toolbus.prepare();
			new Thread(new Runnable(){
				public void run(){
					toolbus.execute();
				}
			}).start();
		}catch(Exception e){
			e.printStackTrace();
			fail(e.getMessage());
		}
	}

	/**
	 * Stops the toolbus.
	 */
	protected void stopToolbus() {
		toolbus.shutdown(factory.makeList());
	}

	public int getPort() {
		return toolbus.getPort();
	}

	public void connectToolComValidator(ToolComValidator toolComValidator)
			throws Exception {
		toolComValidator.connect(toolbus);
	}

	public void addComValidatorForExecutableTool(String name,
			Class<ToolComValidator> toolStubClass) {
		defaultTestToolExecutorFactory.addComValidatorForExecutableTool(name,
				toolStubClass);
	}

	public void testFailed(String message) {
		errorsOccured = true;
		stopToolbus();

		fail(message);
	}

	public boolean hasFailed() {
		return errorsOccured;
	}
}
